import pytest

pytestmark = pytest.mark.slow

version_file_content = """
major = 1
minor = 0
patch = 0
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}.{{patch}}',
}

FILES = ["README.md"]

VERSION = ['major', 'minor', 'patch']
"""


def test_set_part(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 1.0.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    test_environment.call([
        "punch",
        "--action", "punch:set",
        "--action-options", "major=3"
    ])

    assert test_environment.get_file_content("README.md") == "Version 3.0.0"


# def test_set_part_reset_on_set(test_environment):
#     test_environment.ensure_file_is_present("README.md", "Version 1.2.3")

#     test_environment.ensure_file_is_present(
#         "punch_version.py",
#         version_file_content
#     )

#     test_environment.ensure_file_is_present(
#         "punch_config.py",
#         config_file_content
#     )

#     test_environment.call([
#         "punch",
#         "--action", "punch:set",
#         "--action-options", "major=9",
#         "--action-flags", "reset-on-set=true"
#     ])

#     assert test_environment.get_file_content("README.md") == "Version 9.0.0"


def test_punch_set_multiple_version_parts(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 1.0.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        "major = 1\nminor = 0\npatch = 0\n"
    )

    config_file_content = """
    __config_version__ = 1

    GLOBALS = {
        'serializer': '{{major}}.{{minor}}.{{patch}}',
    }

    FILES = ["README.md"]

    VERSION = ['major', 'minor', 'patch']
    """

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    test_environment.call([
        "punch",
        "--action", "punch:set",
        "--action-options", "minor=4,patch=23"
    ])

    assert test_environment.get_file_content("README.md") == "Version 1.4.23"


# def test_punch_set_and_reset_multiple_parts_fails(test_environment):
#     test_environment.ensure_file_is_present("README.md", "Version 1.2.3")

#     test_environment.ensure_file_is_present(
#         "punch_version.py",
#         "major = 1\nminor = 2\npatch = 3\n"
#     )

#     config_file_content = """
#     __config_version__ = 1

#     GLOBALS = {
#         'serializer': '{{major}}.{{minor}}.{{patch}}',
#     }

#     FILES = ["README.md"]

#     VERSION = ['major', 'minor', 'patch']
#     """

#     test_environment.ensure_file_is_present(
#         "punch_config.py",
#         config_file_content
#     )

#     with pytest.raises(subprocess.CalledProcessError):
#         test_environment.output(
#             ["punch", "--set-part", "major=9,minor=8", "--reset-on-set"])



# import pytest

from punch import helpers


def test_optstr2dict_single_option():
    optstr = "part=major"

    assert helpers.optstr2dict(optstr) == {
        'part': 'major'
    }


def test_optstr2dict_multiple_options():
    optstr = "part=major,reset=true"

    assert helpers.optstr2dict(optstr) == {
        'part': 'major',
        'reset': 'true'
    }


def test_optstr2dict_convert_boolean():
    optstr = "reset=true"

    assert helpers.optstr2dict(optstr, convert_boolean=True) == {
        'reset': True
    }


def test_optstr2dict_convert_boolean_false():
    optstr = "reset=false"

    assert helpers.optstr2dict(optstr, convert_boolean=True) == {
        'reset': False
    }


def test_optstr2dict_convert_boolean_mixed_case():
    optstr = "reset=TrUe"

    assert helpers.optstr2dict(optstr, convert_boolean=True) == {
        'reset': True
    }



from __future__ import print_function, absolute_import, division


class IncreasePartAction(object):

    def __init__(self, adict):
        self.part = adict['part']

    def process_version(self, version):
        new_version = version.copy()
        new_version.inc(self.part)

        return new_version



from punch import version
from punch.actions import increase_part


def test_init():
    action_dict = {
        'part': 'somepart'
    }

    action = increase_part.IncreasePartAction(action_dict)

    assert action.part == 'somepart'


def test_process_version():
    v = version.Version()
    v.create_part('major', 4)
    v.create_part('minor', 3)
    v.create_part('patch', 1)

    action_dict = {
        'part': 'major'
    }

    action = increase_part.IncreasePartAction(action_dict)

    new_version = action.process_version(v)

    assert new_version.get_part('major').value == 5
    assert new_version.get_part('minor').value == 0
    assert new_version.get_part('patch').value == 0



import subprocess

import os
import pytest

import sys
from six.moves import configparser

from punch import vcs_configuration as vc
from punch.vcs_repositories import hg_repo as hr, exceptions as re
from tests.conftest import safe_devnull


pytestmark = pytest.mark.slow


def hg_repo_add_file(temp_hg_dir, fname, content="", out=None):
    if out is None:
        out = safe_devnull()

    with open(os.path.join(temp_hg_dir, fname), "w") as f:
        f.write(content)

    subprocess.check_call(["hg", "add", fname], cwd=temp_hg_dir, stdout=out)


def hg_repo_add_branch(temp_hg_dir, branch, message=None, out=None):
    if out is None:
        out = sys.stdout
    if message is None:
        message = "Starting new branch " + branch
    subprocess.check_call(
        ["hg", "branch", "-f", branch],
        cwd=temp_hg_dir,
        stdout=out
    )
    subprocess.check_call(
        ["hg", "commit", "-m", message],
        cwd=temp_hg_dir,
        stdout=out
    )


def hg_repo_change_branch(temp_hg_dir, branch, out=None):
    if out is None:
        out = sys.stdout
    subprocess.check_call(
        ["hg", "update", branch],
        cwd=temp_hg_dir,
        stdout=out
    )


def hg_log(temp_hg_dir):
    p = subprocess.Popen(["hg", "log"], cwd=temp_hg_dir,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    return stdout.decode('utf8')


@pytest.fixture
def temp_empty_hg_dir(temp_empty_dir):
    subprocess.check_call(["hg", "init", temp_empty_dir])

    hgrc = os.path.join(temp_empty_dir, ".hg", "hgrc")
    cp = configparser.ConfigParser()
    cp.read(hgrc)
    cp.add_section("ui")
    cp.set("ui", "username", "PyTest <py.test@email.com>")

    with open(hgrc, "w+") as f:
        cp.write(f)

    return temp_empty_dir


@pytest.fixture
def temp_hg_dir(temp_empty_hg_dir, safe_devnull):
    with open(os.path.join(temp_empty_hg_dir, "README.md"), "w") as f:
        f.writelines(["# Test file", "This is just a test file for punch"])

    subprocess.check_call(["hg", "add", "README.md"],
                          cwd=temp_empty_hg_dir, stdout=safe_devnull)
    subprocess.check_call(
        ["hg", "commit", "-m", "Initial addition"],
        cwd=temp_empty_hg_dir,
        stdout=safe_devnull
    )

    return temp_empty_hg_dir


@pytest.fixture
def hg_other_branch(temp_hg_dir):
    hg_repo_add_branch(temp_hg_dir, "other")
    hg_repo_change_branch(temp_hg_dir, hr.HgRepo.DEFAULT_BRANCH)
    return temp_hg_dir


@pytest.fixture
def empty_vcs_configuration():
    return vc.VCSConfiguration(
        'hg', {}, {}, {'current_version': 'a', 'new_version': 'b'}
    )


@pytest.fixture
def other_branch_vcs_configuration():
    return vc.VCSConfiguration(
        'hg', {"branch": "other"}, {},
        {'current_version': 'a', 'new_version': 'b'}
    )


@pytest.fixture
def ready_to_finish_repo(temp_hg_dir, **kwargs):
    release_name = "1.0"
    commit_message = "A commit message"
    config = vc.VCSConfiguration(
        'git', kwargs, global_variables={},
        special_variables={'new_version': release_name},
        commit_message=commit_message
    )

    repo = hr.HgRepo(temp_hg_dir, config)
    repo.pre_start_release()
    repo.start_release()

    hg_repo_add_file(temp_hg_dir, "version.txt", release_name + "\n")

    return repo


def test_init(temp_empty_hg_dir, empty_vcs_configuration):
    repo = hr.HgRepo(temp_empty_hg_dir, empty_vcs_configuration)
    assert repo.working_path == temp_empty_hg_dir


def test_init_with_uninitialized_dir(temp_empty_dir, empty_vcs_configuration):
    with pytest.raises(re.RepositorySystemError) as exc:
        hr.HgRepo(temp_empty_dir, empty_vcs_configuration)

    assert str(exc.value) == \
        "The current directory {} is not a Hg repository".format(
            temp_empty_dir)


def test_get_current_branch(temp_hg_dir, empty_vcs_configuration):
    repo = hr.HgRepo(temp_hg_dir, empty_vcs_configuration)
    assert repo.get_current_branch() == repo.DEFAULT_BRANCH


def test_get_tags(temp_hg_dir, empty_vcs_configuration):
    repo = hr.HgRepo(temp_hg_dir, empty_vcs_configuration)
    assert repo.get_tags() == 'tip'


def test_pre_start_release(temp_hg_dir, empty_vcs_configuration):
    repo = hr.HgRepo(temp_hg_dir, empty_vcs_configuration)
    repo.pre_start_release()

    assert repo.get_current_branch() == repo.DEFAULT_BRANCH


def test_pre_start_release_start_from_different_branch(
        temp_hg_dir, empty_vcs_configuration):
    hg_dir = temp_hg_dir
    repo = hr.HgRepo(hg_dir, empty_vcs_configuration)
    hg_repo_add_branch(hg_dir, "other")
    repo.pre_start_release()
    assert repo.get_current_branch() == repo.DEFAULT_BRANCH


def test_pre_start_release_should_use_other_branch(
        temp_hg_dir, other_branch_vcs_configuration):
    hg_dir = temp_hg_dir
    repo = hr.HgRepo(hg_dir, other_branch_vcs_configuration)
    hg_repo_add_branch(hg_dir, "other")
    repo.pre_start_release()
    hg_repo_change_branch(hg_dir, repo.DEFAULT_BRANCH)
    repo.pre_start_release()
    assert repo.get_current_branch() == "other"


def test_pre_start_release_in_unclean_state(
        temp_hg_dir, empty_vcs_configuration):
    hg_dir = temp_hg_dir
    with open(os.path.join(hg_dir, "README.md"), "w") as f:
        f.writelines(["Uncommitted lines"])
    repo = hr.HgRepo(hg_dir, empty_vcs_configuration)
    with pytest.raises(re.RepositoryStatusError):
        repo.pre_start_release()

    assert repo.get_current_branch() == repo.DEFAULT_BRANCH


def test_pre_start_release_starting_from_different_branch_in_unclean_state(
        temp_hg_dir, empty_vcs_configuration):
    hg_dir = temp_hg_dir
    hg_repo_add_branch(hg_dir, "other")
    with open(os.path.join(hg_dir, "README.md"), "w") as f:
        f.writelines(["Uncommitted lines"])
    repo = hr.HgRepo(hg_dir, empty_vcs_configuration)
    with pytest.raises(re.RepositoryStatusError):
        repo.pre_start_release()

    assert repo.get_current_branch() == "other"


def test_start_release_should_be_in_defined_branch(
        hg_other_branch, other_branch_vcs_configuration):
    repo = hr.HgRepo(hg_other_branch, other_branch_vcs_configuration)
    repo.pre_start_release()
    repo.start_release()
    assert repo.get_current_branch() == "other"


def test_finish_release_without_changes(
        hg_other_branch, other_branch_vcs_configuration):
    repo = hr.HgRepo(hg_other_branch, other_branch_vcs_configuration)
    repo.pre_start_release()
    repo.start_release()
    repo.finish_release()
    assert repo.get_current_branch() == repo.DEFAULT_BRANCH
    assert 'b' not in repo.get_tags()


def test_finish_should_recover_start_branch(
        hg_other_branch, other_branch_vcs_configuration):
    hg_repo_add_branch(hg_other_branch, "third")
    repo = hr.HgRepo(hg_other_branch, other_branch_vcs_configuration)
    repo.pre_start_release()
    repo.start_release()
    repo.finish_release()
    assert repo.get_current_branch() == "third"


def test_finish_release_with_message(ready_to_finish_repo):
    d = ready_to_finish_repo.working_path
    config = ready_to_finish_repo.config_obj
    commit_message = config.commit_message

    ready_to_finish_repo.finish_release()

    log = hg_log(d)
    assert commit_message in log


def test_finish_release_without_release_branch(ready_to_finish_repo):
    config = ready_to_finish_repo.config_obj
    release_name = config.options['new_version']

    ready_to_finish_repo.finish_release()

    assert release_name not in ready_to_finish_repo.get_branches()


def test_finish_write_tag(ready_to_finish_repo):
    config = ready_to_finish_repo.config_obj
    release_name = config.options['new_version']

    ready_to_finish_repo.finish_release()

    assert release_name in ready_to_finish_repo.get_tags()


def test_finish_release_with_custom_tag(temp_hg_dir):
    tag = "Version_{}".format("1.0")
    repo = ready_to_finish_repo(temp_hg_dir, tag=tag)
    repo.finish_release()

    assert tag in repo.get_tags()


def test_finish_release_custom_tag_cannot_contain_spaces(temp_hg_dir):
    release_name = "1.0"
    commit_message = "A commit message"
    tag = "Version {}".format(release_name)

    config = vc.VCSConfiguration('hg', {'tag': tag}, global_variables={},
                                 special_variables={
                                     'new_version': release_name},
                                 commit_message=commit_message)

    with pytest.raises(re.RepositoryConfigurationError):
        hr.HgRepo(temp_hg_dir, config)


def test_finish_release_custom_tag_cannot_be_a_number(temp_hg_dir):
    release_name = "1.0"
    commit_message = "A commit message"
    tag = "12234"

    config = vc.VCSConfiguration('hg', {'tag': tag}, global_variables={},
                                 special_variables={
                                     'new_version': release_name},
                                 commit_message=commit_message)

    with pytest.raises(re.RepositoryConfigurationError):
        hr.HgRepo(temp_hg_dir, config)


def test_start_summary(temp_hg_dir, empty_vcs_configuration):
    repo = hr.HgRepo(temp_hg_dir, empty_vcs_configuration)
    repo.pre_start_release()
    assert repo.get_summary() == {"branch": "default", "commit": "(clean)",
                                  "update": "(current)"}


def test_get_branches(hg_other_branch, empty_vcs_configuration):
    repo = hr.HgRepo(hg_other_branch, empty_vcs_configuration)
    assert {"default", "other"} == repo.get_branches()


def test_tag(temp_hg_dir, empty_vcs_configuration):
    repo = hr.HgRepo(temp_hg_dir, empty_vcs_configuration)

    repo.tag("just_a_tag")

    assert "just_a_tag" in repo.get_tags()



import pytest
import six
from punch.vcs_repositories import vcs_repo as vr, exceptions as re

if six.PY2:
    import mock
else:
    from unittest import mock

pytestmark = pytest.mark.slow


def _test_set_command(self):
    self.commands = ['ls']
    self.command = 'ls'


def test_init_without_program_installed(temp_empty_dir):
    with mock.patch("subprocess.Popen") as mock_popen:
        if six.PY2:
            mock_popen.side_effect = IOError
        else:
            mock_popen.side_effect = FileNotFoundError

        with pytest.raises(re.RepositorySystemError):
            vr.VCSRepo._set_command = _test_set_command
            vr.VCSRepo(temp_empty_dir, mock.Mock())


def test_run_without_errors(temp_empty_dir):
    with mock.patch('subprocess.check_call'):
        vr.VCSRepo._set_command = _test_set_command
        repo = vr.VCSRepo(temp_empty_dir, mock.Mock())

    with mock.patch('subprocess.Popen') as mock_popen:
        mock_popen_obj = mock.Mock()
        mock_popen.return_value = mock_popen_obj
        mock_popen_obj.communicate.return_value = \
            ("stdout".encode("utf8"), "stderr".encode("utf8"))
        mock_popen_obj.returncode = 0

        assert repo._run([repo.command, "--help"]) == "stdout"


def test_run_with_errors(temp_empty_dir):
    with mock.patch('subprocess.check_call'):
        vr.VCSRepo._set_command = _test_set_command
        repo = vr.VCSRepo(temp_empty_dir, mock.Mock())

    with mock.patch('subprocess.Popen') as mock_popen:
        mock_popen_obj = mock.Mock()
        mock_popen.return_value = mock_popen_obj
        mock_popen_obj.communicate.return_value = \
            ("stdout".encode("utf8"), "stderr".encode("utf8"))
        mock_popen_obj.returncode = 1

        with pytest.raises(vr.RepositorySystemError) as exc:
            assert repo._run([repo.command, "--help"])

        assert str(exc.value) == \
            "An error occurred executing '{} --help':".format(repo.command) + \
            " stderr\nProcess output was: stdout"


def test_initialize_repo_with_global_configuration_object(
        mocker, temp_empty_dir):
    global_config = mocker.Mock()
    repo = vr.VCSRepo(temp_empty_dir, config_obj=global_config)
    assert repo.config_obj == global_config



from __future__ import print_function, absolute_import, division

import os
import six

from punch import replacer


class FileUpdater(object):

    def __init__(self, file_configuration):
        self.file_configuration = file_configuration
        self.rep = replacer.Replacer(file_configuration.config['serializer'])

    def get_summary(self, current_version, new_version):
        return self.rep.run_all_serializers(current_version, new_version)

    def update(self, current_version, new_version):
        if not os.path.exists(self.file_configuration.path):
            if six.PY2:
                raise IOError(
                    "The file {} does not exist".format(
                        self.file_configuration.path
                    )
                )
            else:
                raise FileNotFoundError(
                    "The file {} does not exist".format(
                        self.file_configuration.path
                    )
                )

        with open(self.file_configuration.path, 'r') as f:
            old_file_content = f.read()

        new_file_content = self.rep.replace(
            old_file_content,
            current_version,
            new_version
        )

        if six.PY2:
            new_file_content = new_file_content.encode('utf8')

        if new_file_content == old_file_content:
            raise ValueError(
                "Cannot find any match for version {} in file {}".format(
                    current_version, self.file_configuration.path
                )
            )

        with open(self.file_configuration.path, 'w') as f:
            f.write(new_file_content)



import os
import subprocess

import pytest

import punch.version as ver
import punch.version_part as vpart

pytestmark = pytest.mark.slow

version_file_content = """
major = '2016'
minor = '4'
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}',
}

FILES = ["README.md"]

VERSION = [
    {
        'name': 'major',
        'type': 'date',
        'fmt': 'YYYY'
    },
    {
        'name': 'minor',
        'type': 'date',
        'fmt': 'MM'
    }
]
"""


@pytest.fixture
def version_mm():
    v = ver.Version()
    v.create_part('major', '2018', vpart.DateVersionPart, '%Y')
    v.create_part('minor', '04', vpart.DateVersionPart, '%m')
    return v


def clean_previous_imports():
    import sys

    for i in ['punch_config', 'punch_version']:
        if i in sys.modules:
            sys.modules.pop(i)


def test_update_major(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 2016.4.")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    system_year = subprocess.check_output(['date', '+%Y'])
    system_year = system_year.decode('utf8').replace('\n', '')

    system_month = subprocess.check_output(['date', '+%m'])
    system_month = system_month.decode('utf8').\
        replace('\n', '').lstrip('0')

    test_environment.call(["punch", "--part", "major"])

    assert test_environment.get_file_content("README.md") == \
        "Version {}.{}.".format(system_year, system_month)


def test_update_minor(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 2016.4.")

    test_environment.ensure_file_is_present(
        "punch_version.py", version_file_content)

    test_environment.ensure_file_is_present(
        "punch_config.py", config_file_content)

    system_month = subprocess.check_output(['date', '+%m'])
    system_month = system_month.decode(
        'utf8').replace('\n', '').lstrip('0')

    test_environment.call(["punch", "--part", "minor"])

    assert test_environment.get_file_content("README.md") == \
        "Version 2016.{}.".format(system_month)


def test_write_version_file(temp_empty_dir, version_mm):
    clean_previous_imports()

    version_filepath = os.path.join(temp_empty_dir, 'punch_version.py')

    version_mm.to_file(version_filepath)

    with open(version_filepath, 'r') as f:
        content = sorted(f.readlines())

    # Uncomment to get more insight into test failures.
    # print(content)

    expected_content = [
        "major = '2018'\n",
        "minor = '04'\n",
    ]

    assert content == expected_content



from __future__ import print_function, absolute_import, division


class VCSUseCase(object):
    def __init__(self, repo):
        self.repo = repo

    def __getattr__(self, item):
        return getattr(self.repo, item)



#!/usr/bin/env python

from __future__ import print_function, absolute_import, division

import argparse
import sys

import os
import punch
from punch import config as cfr
from punch import file_updater as fu
from punch import replacer as rep
from punch import vcs_configuration as vcsc
from punch import version as ver
from punch import action_register as ar
from punch import helpers as hlp
from punch.vcs_repositories import exceptions as rex
from punch.vcs_repositories import git_flow_repo as gfr
from punch.vcs_repositories import git_repo as gr
from punch.vcs_use_cases import release as ruc


def fatal_error(message, exception=None):
    print(message)
    if exception is not None:
        print("Exception {}: {}".format(
            exception.__class__.__name__,
            str(exception)
        ))
    sys.exit(1)


default_config_file_name = "punch_config.py"

default_config_file_content = """__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}.{{patch}}',
}

FILES = []

VERSION = ['major', 'minor', 'patch']

VCS = {
    'name': 'git',
    'commit_message': (
        "Version updated from {{ current_version }}",
        " to {{ new_version }}"),
}
"""

default_version_file_name = "punch_version.py"

default_version_file_content = """major = 0
minor = 1
patch = 0
"""

default_commit_message = \
    "Version update {{ current_version }} -> {{ new_version }}"


def show_version_parts(values):
    for p in values:
        print("{}={}".format(p.name, p.value))


def show_version_updates(version_changes):
    for current, new in version_changes:
        print("  * {} -> {}".format(current, new))


def main(original_args=None):
    parser = argparse.ArgumentParser(
        description="Manages file content with versions."
    )
    parser.add_argument('-c', '--config-file', action='store',
                        help="Config file", default=default_config_file_name)
    parser.add_argument('-v', '--version-file', action='store',
                        help="Version file", default=default_version_file_name)
    parser.add_argument('-p', '--part', action='store')
    parser.add_argument('--set-part', action='store')
    parser.add_argument('-a', '--action', action='store')
    parser.add_argument('--action-options', action='store')
    parser.add_argument('--action-flags', action='store')
    parser.add_argument('--reset-on-set', action='store_true')
    parser.add_argument('--verbose', action='store_true',
                        help="Be verbose")
    parser.add_argument('--version', action='store_true',
                        help="Print the Punch version and project information")
    parser.add_argument(
        '--init',
        action='store_true',
        help="Writes default initialization files" +
             " (does not overwrite existing ones)"
    )
    parser.add_argument(
        '-s',
        '--simulate',
        action='store_true',
        help="Simulates the version increment and" +
             " prints a summary of the relevant data (implies --verbose)"
    )

    args = parser.parse_args()

    # These are here just to avoid "can be not defined" messages by linters
    config = None
    repo = None

    if args.version is True:
        print("Punch version {}".format(punch.__version__))
        print("Copyright (C) 2016 Leonardo Giordani")
        print("This is free software, see the LICENSE file.")
        print("Source: https://github.com/lgiordani/punch")
        print("Documentation: http://punch.readthedocs.io/en/latest/")
        sys.exit(0)

    if args.simulate:
        args.verbose = True

    if args.init is True:
        if not os.path.exists(default_config_file_name):
            with open(default_config_file_name, 'w') as f:
                f.write(default_config_file_content)

        if not os.path.exists(default_version_file_name):
            with open(default_version_file_name, 'w') as f:
                f.write(default_version_file_content)

        sys.exit(0)

    if not any([args.part, args.set_part, args.action]):
        fatal_error("You must specify one of --part, --set-part, or --action")

    if args.set_part and args.reset_on_set:
        set_parts = args.set_part.split(',')
        if len(set_parts) > 1:
            fatal_error(
                "If you specify --reset-on-set you may set only one value"
            )

    if args.verbose:
        print("## Punch version {}".format(punch.__version__))

    try:
        config = cfr.PunchConfig(args.config_file)
    except (cfr.ConfigurationVersionError, ValueError) as exc:
        fatal_error(
            "An error occurred while reading the configuration file.",
            exc
        )

    if not args.simulate:
        if len(config.files) == 0:
            fatal_error("You didn't configure any file")

    current_version = ver.Version.from_file(args.version_file, config.version)
    new_version = current_version.copy()

    if args.part:
        args.action = "punch:increase"
        args.action_options = "part={}".format(args.part)

    if args.set_part:
        args.action = "punch:set"
        args.action_options = args.set_part

    if args.action:
        try:
            action_dict = config.actions[args.action]
        except KeyError:
            print(
                "The requested action {} is not defined.".format(
                    args.action
                )
            )
            sys.exit(0)

        try:
            action_name = action_dict.pop('type')
        except KeyError:
            print("The action configuration is missing the 'type' field.")
            sys.exit(0)

        if args.action_options:
            action_dict.update(hlp.optstr2dict(args.action_options))

        action_class = ar.ActionRegister.get(action_name)
        action = action_class(action_dict)
        new_version = action.process_version(new_version)

    global_replacer = rep.Replacer(config.globals['serializer'])
    current_version_string, new_version_string = \
        global_replacer.run_main_serializer(
            current_version.as_dict(),
            new_version.as_dict()
        )

    if config.vcs is not None:
        special_variables = {
            'current_version': current_version_string,
            'new_version': new_version_string
        }

        vcs_configuration = vcsc.VCSConfiguration.from_dict(
            config.vcs,
            config.globals,
            special_variables
        )
    else:
        vcs_configuration = None

    if args.verbose:
        print("\n* Current version")
        show_version_parts(current_version.values)

        print("\n* New version")
        show_version_parts(new_version.values)

        changes = global_replacer.run_all_serializers(
            current_version.as_dict(),
            new_version.as_dict()
        )

        print("\n* Global version updates")
        show_version_updates(changes)

        print("\nConfigured files")
        for file_configuration in config.files:
            updater = fu.FileUpdater(file_configuration)
            print("* {}:".format(file_configuration.path))
            changes = updater.get_summary(
                current_version.as_dict(),
                new_version.as_dict()
            )
            show_version_updates(changes)

        if vcs_configuration is not None:
            print("\nVersion control configuration")
            print("Name:", vcs_configuration.name)
            print("Commit message", vcs_configuration.commit_message)
            print("Options:", vcs_configuration.options)

    if not args.simulate:
        if vcs_configuration is not None:
            if vcs_configuration.name == 'git':
                repo_class = gr.GitRepo
            elif vcs_configuration.name == 'git-flow':
                repo_class = gfr.GitFlowRepo
            else:
                fatal_error(
                    "The requested version control" +
                    " system {} is not supported.".format(
                        vcs_configuration.name
                    )
                )

            try:
                repo = repo_class(os.getcwd(), vcs_configuration)
            except rex.RepositorySystemError as exc:
                fatal_error(
                    "An error occurred while initializing" +
                    " the version control repository",
                    exc
                )
        else:
            repo = None

        if vcs_configuration is not None:
            # TODO: Create a fake UseCase to allow running this
            # without a repo and outside this nasty if
            uc = ruc.VCSReleaseUseCase(repo)
            uc.pre_start_release()
            uc.start_release()

        for file_configuration in config.files:
            updater = fu.FileUpdater(file_configuration)
            try:
                updater.update(
                    current_version.as_dict(), new_version.as_dict()
                )
            except ValueError as e:
                print("Warning:", e)

        with open(args.version_file, 'w') as f:
            for i in new_version.keys:
                f.write('{name} = {value}\n'.format(
                    name=new_version.parts[i].name,
                    value=new_version.parts[i].value
                ))

        if vcs_configuration is not None:
            uc.finish_release()
            uc.post_finish_release()



import pytest

pytestmark = pytest.mark.slow

version_file_content = """
major = 1
minor = 0
patch = 0
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}.{{patch}}',
}

FILES = ["README.md"]

VERSION = ['major', 'minor', 'patch']
"""


def test_action_increase(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 1.0.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    test_environment.call([
        "punch",
        "--action", "punch:increase",
        "--action-options", "part=major"
    ])

    assert test_environment.get_file_content("README.md") == "Version 2.0.0"


def test_action_increase_resets_fields(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 1.1.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        """
        major = 1
        minor = 1
        patch = 0
        """
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    test_environment.call([
        "punch",
        "--action", "punch:increase",
        "--action-options", "part=major"
    ])

    assert test_environment.get_file_content("README.md") == "Version 2.0.0"


def test_option_part(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 1.0.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    test_environment.call([
        "punch",
        "--part", "major"
    ])

    assert test_environment.get_file_content("README.md") == "Version 2.0.0"



from punch import version
from punch.actions import set_part


def test_init():
    action_dict = {
        'minor': '4'
    }

    action = set_part.SetPartAction(action_dict)

    assert action.parts == action_dict


def test_process_version():
    v = version.Version()
    v.create_part('major', 4)
    v.create_part('minor', 3)
    v.create_part('patch', 1)

    action_dict = {
        'minor': '4'
    }

    action = set_part.SetPartAction(action_dict)

    new_version = action.process_version(v)

    assert new_version.get_part('major').value == 4
    assert new_version.get_part('minor').value == 4
    assert new_version.get_part('patch').value == 1



from __future__ import print_function, absolute_import, division

import os
import six
from punch.vcs_repositories import exceptions as e
from punch.vcs_repositories import vcs_repo as vr


class GitRepo(vr.VCSRepo):

    def __init__(self, working_path, config_obj):
        if six.PY2:
            super(GitRepo, self).__init__(working_path, config_obj)
        else:
            super().__init__(working_path, config_obj)

        self.make_release_branch = self.config_obj.options.get(
            'make_release_branch',
            True
        )

    def _check_config(self):
        # Tag names cannot contain spaces
        tag = self.config_obj.options.get('tag', '')
        if ' ' in tag:
            raise e.RepositoryConfigurationError(
                """You specified "'tag': {}".""".format(tag) +
                "Tag names cannot contain spaces")

    def _check_system(self):
        if six.PY2:
            super(GitRepo, self)._check_system()
        else:
            super()._check_system()

        if not os.path.exists(os.path.join(self.working_path, '.git')):
            raise e.RepositorySystemError(
                "The current directory {}".format(self.working_path) +
                " is not a Git repository")

    def _set_command(self):
        self.commands = ['git']
        self.command = 'git'

    def get_current_branch(self):
        stdout = self._run([self.command, "rev-parse", "--abbrev-ref", "HEAD"])

        branch = stdout.replace("\n", "")

        return branch

    def get_tags(self):
        return self._run([self.command, "tag"])

    def get_branches(self):
        return self._run([self.command, "branch"])

    def pre_start_release(self):
        output = self._run([self.command, "status"])
        if "Changes to be committed:" in output:
            raise e.RepositoryStatusError(
                "Cannot checkout master while repository" +
                " contains uncommitted changes")

        self._run([self.command, "checkout", "master"])

        branch = self.get_current_branch()

        if branch != "master":
            raise e.RepositoryStatusError(
                "Current branch shall be master but is {}".format(branch))

    def start_release(self):
        if self.make_release_branch:
            self._run([
                self.command,
                "checkout",
                "-b",
                self.config_obj.options['new_version']
            ])

    def finish_release(self):
        branch = self.get_current_branch()

        # self._run([self.command, "add", "."])

        output = self._run([self.command, "status"])

        if ("nothing to commit, working directory clean" in output or
                "nothing to commit, working tree clean" in output) and \
                self.make_release_branch:
            self._run([self.command, "checkout", "master"])
            self._run([self.command, "branch", "-d", branch])
            return

        command_line = [self.command, "commit"]
        command_line.extend(["-m", self.config_obj.commit_message])

        self._run(command_line)

        if self.make_release_branch:
            self._run([self.command, "checkout", "master"])
            self._run([self.command, "merge", branch])
            self._run([self.command, "branch", "-d", branch])

        try:
            tag_value = self.config_obj.options['tag']
        except KeyError:
            tag_value = self.config_obj.options['new_version']

        if self.config_obj.options.get('annotate_tags', False):
            annotation_message = self.config_obj.options.get(
                'annotation_message', "Version {{ new_version }}")
            self._run([
                self.command,
                "tag",
                "-a",
                tag_value,
                "-m",
                annotation_message
            ])
        else:
            self._run([self.command, "tag", tag_value])

    def post_finish_release(self):
        pass

    def tag(self, tag_name):
        self._run([self.command, "tag", tag_name])



from __future__ import print_function, absolute_import, division

from punch.actions import increase_part
from punch.actions import set_part
from punch.actions import conditional_reset


class ActionRegister(object):
    _register = {}

    @classmethod
    def register(cls, aclass, class_name):
        cls._register[class_name] = aclass

    @classmethod
    def get(cls, action_name):
        return cls._register[action_name]


ActionRegister.register(increase_part.IncreasePartAction, 'increase_part')
ActionRegister.register(set_part.SetPartAction, 'set_part')
ActionRegister.register(
    conditional_reset.ConditionalResetAction, 'conditional_reset'
)



import pytest

import punch

pytestmark = pytest.mark.slow

version_file_content = """
major = 1
minor = 0
patch = 0
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}.{{patch}}',
}

FILES = ["README.md"]

VERSION = ['major', 'minor', 'patch']
"""


def test_verbose(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 1.0.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    ret = test_environment.call(["punch", "--verbose", "--part", "major"])

    assert test_environment.get_file_content("README.md") == "Version 2.0.0"
    assert not ret.stderr
    assert ret.stdout == """## Punch version {version}

* Current version
major=1
minor=0
patch=0

* New version
major=2
minor=0
patch=0

* Global version updates
  * 1.0.0 -> 2.0.0

Configured files
* README.md:
  * 1.0.0 -> 2.0.0
""".format(version=punch.__version__)


def test_simulate(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 1.0.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    ret = test_environment.call([
        "punch",
        "--simulate",
        "--part",
        "major"
    ])

    assert test_environment.get_file_content("README.md") == "Version 1.0.0"
    assert not ret.stderr
    assert ret.stdout == """## Punch version {version}

* Current version
major=1
minor=0
patch=0

* New version
major=2
minor=0
patch=0

* Global version updates
  * 1.0.0 -> 2.0.0

Configured files
* README.md:
  * 1.0.0 -> 2.0.0
""".format(version=punch.__version__)


def test_simulate_and_verbose(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 1.0.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    ret = test_environment.call([
        "punch",
        "--simulate",
        "--verbose",
        "--part",
        "major"
    ])

    assert test_environment.get_file_content("README.md") == "Version 1.0.0"
    assert not ret.stderr
    assert ret.stdout == """## Punch version {version}

* Current version
major=1
minor=0
patch=0

* New version
major=2
minor=0
patch=0

* Global version updates
  * 1.0.0 -> 2.0.0

Configured files
* README.md:
  * 1.0.0 -> 2.0.0
""".format(version=punch.__version__)



import six

import pytest
import io

from punch import replacer


def file_like(file_content):
    if six.PY2:
        return io.StringIO(unicode(file_content)) # NOQA
    else:
        return io.StringIO(file_content)


def test_replace_content_without_config():
    with pytest.raises(TypeError):
        replacer.Replacer()


def test_replace_content():
    current_version = {
        'major': 1,
        'minor': 0,
        'patch': 0
    }
    new_version = {
        'major': 1,
        'minor': 0,
        'patch': 1
    }

    file_content = """# Just a comment
    __version__ = "1.0.0"
    """

    updated_file_content = """# Just a comment
    __version__ = "1.0.1"
    """

    serializer = "__version__ = \"{{major}}.{{minor}}.{{patch}}\""
    rep = replacer.Replacer(serializer)

    new_file_content = rep.replace(file_content, current_version, new_version)

    assert new_file_content == updated_file_content


def test_get_versions():
    current_version = {
        'major': 1,
        'minor': 0,
        'patch': 0
    }
    new_version = {
        'major': 1,
        'minor': 0,
        'patch': 1
    }

    serializer = "__version__ = \"{{major}}.{{minor}}.{{patch}}\""
    rep = replacer.Replacer(serializer)

    list_of_versions = rep.run_all_serializers(current_version, new_version)

    assert list_of_versions == [
        ("__version__ = \"1.0.0\"", "__version__ = \"1.0.1\"")
    ]


def test_get_versions_with_multiple_serializers():
    current_version = {
        'major': 1,
        'minor': 0,
        'patch': 0
    }
    new_version = {
        'major': 1,
        'minor': 0,
        'patch': 1
    }

    serializers = [
        "__version__ = \"{{major}}.{{minor}}.{{patch}}\"",
        "__api_abi__ = \"{{major}}.{{minor}}\""
    ]
    rep = replacer.Replacer(serializers)

    list_of_versions = rep.run_all_serializers(current_version, new_version)

    assert list_of_versions == [
        ("__version__ = \"1.0.0\"", "__version__ = \"1.0.1\""),
        ("__api_abi__ = \"1.0\"", "__api_abi__ = \"1.0\"")
    ]


def test_get_main_version_change_with_multiple_serializers():
    current_version = {
        'major': 1,
        'minor': 0,
        'patch': 0
    }
    new_version = {
        'major': 1,
        'minor': 0,
        'patch': 1
    }

    serializers = [
        "__version__ = \"{{major}}.{{minor}}.{{patch}}\"",
        "__api_abi__ = \"{{major}}.{{minor}}\""
    ]
    rep = replacer.Replacer(serializers)

    current, new = rep.run_main_serializer(current_version, new_version)

    assert current, new == (
        "__version__ = \"1.0.0\"", "__version__ = \"1.0.1\""
    )


def test_replace_content_with_multiple_serializers():
    current_version = {
        'major': 1,
        'minor': 0,
        'patch': 0
    }
    new_version = {
        'major': 1,
        'minor': 0,
        'patch': 1
    }

    file_content = """# Just a comment
    __version__ = "1.0.0"
    __api_abi__ = "1.0"
    """

    updated_file_content = """# Just a comment
    __version__ = "1.0.1"
    __api_abi__ = "1.0"
    """

    serializers = [
        "__version__ = \"{{major}}.{{minor}}.{{patch}}\"",
        "__api_abi__ = \"{{major}}.{{minor}}\""
    ]

    rep = replacer.Replacer(serializers)

    new_file_content = rep.replace(file_content, current_version, new_version)

    assert new_file_content == updated_file_content


def test_replace_content_without_using_all_parts():
    current_version = {
        'major': 1,
        'minor': 0,
        'patch': 0
    }
    new_version = {
        'major': 1,
        'minor': 1,
        'patch': 0
    }

    file_content = """# Just a comment
    __version__ = "1.0"
    """

    updated_file_content = """# Just a comment
    __version__ = "1.1"
    """

    serializer = "__version__ = \"{{major}}.{{minor}}\""
    rep = replacer.Replacer(serializer)

    new_file_content = rep.replace(file_content, current_version, new_version)

    assert new_file_content == updated_file_content



import subprocess
import pytest

pytestmark = pytest.mark.slow

version_file_content = """
year = '2016'
month = '04'
build = 1
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{year}}.{{month}}.{{build}}',
}

FILES = ["README.md"]

ACTIONS = {
    'mbuild': {
        'type': 'conditional_reset',
        'field': 'build',
        'update_fields': ['year', 'month']
    }
}

VERSION = [
    {
        'name': 'year',
        'type': 'date',
        'fmt': '%Y'
    },
    {
        'name': 'month',
        'type': 'date',
        'fmt': '%m'
    },
    'build'
]
"""


def test_action_conditional_reset(script_runner, test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 2016.04.1.")

    test_environment.ensure_file_is_present("punch_version.py",
                                            version_file_content)

    test_environment.ensure_file_is_present("punch_config.py",
                                            config_file_content)

    system_year = subprocess.check_output(['date', '+%Y'])
    system_year = system_year.decode('utf8').replace('\n', '')

    system_month = subprocess.check_output(['date', '+%m'])
    system_month = system_month.decode('utf8').replace('\n', '')

    ret = test_environment.call(['punch', '--action', 'mbuild'])

    assert not ret.stderr
    assert test_environment.get_file_content("README.md") == \
        "Version {}.{}.0.".format(system_year, system_month)



import pytest

pytestmark = pytest.mark.slow

version_file_content = """
major = 0
minor = 2
patch = 0
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}.{{patch}}',
}

FILES = ["VERSION"]

VERSION = ['major', 'minor', 'patch']

VCS = {
    'name': 'git',
}
"""


def test_update_major(test_environment):
    test_environment.ensure_file_is_present("VERSION", "0.2.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    test_environment.output(["git", "init"])

    test_environment.output(["git", "add", "punch_config.py"])

    test_environment.output(["git", "commit", "-m", "some message"])

    test_environment.ensure_file_is_present("untracked_file")

    test_environment.call(["punch", "--part", "minor"])

    out = test_environment.output(
        ["git", "ls-tree", "-r", "master", "--name-only"]
    )

    assert "untracked_file" not in out



import pytest

pytestmark = pytest.mark.slow

version_file_content = """
major = 0
minor = 2
patch = 0
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}.{{patch}}',
}

FILES = ["VERSION"]

VERSION = ['major', 'minor', 'patch']
"""


def test_update_major(test_environment):
    # This is not matching the version_file_content on purpose
    test_environment.ensure_file_is_present("VERSION", "0.1.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    out = test_environment.output(["punch", "--part", "minor"])

    assert test_environment.get_file_content("VERSION") == "0.1.0"
    assert "Warning" in out
    assert "VERSION" in out



import pytest

from punch import version as ver
from punch import version_part as vp
from punch.actions import conditional_reset


def test_init():
    action_dict = {
        'field': 'afield'
    }

    action = conditional_reset.ConditionalResetAction(action_dict)

    assert action.field == 'afield'
    assert action.update_fields is None


def test_init_with_update_fields():
    action_dict = {
        'field': 'afield',
        'update_fields': ['field1', 'field2']
    }

    action = conditional_reset.ConditionalResetAction(action_dict)

    assert action.field == 'afield'
    assert action.update_fields == ['field1', 'field2']


def test_conditional_reset_init_with_no_field():
    with pytest.raises(KeyError):
        conditional_reset.ConditionalResetAction({})


def test_conditional_reset_process_version_checks_all_update_fields(mocker):
    mocker.patch('punch.version_part.DateVersionPart.inc')
    v = ver.Version()
    part_year = vp.DateVersionPart('year', 2016, '%Y')
    part_month = vp.DateVersionPart('month', 1, '%m')
    part_build = vp.IntegerVersionPart('build')
    v.add_part(part_year)
    v.add_part(part_month)
    v.add_part(part_build)

    a = conditional_reset.ConditionalResetAction({
        'field': 'build',
        'update_fields': ['year', 'month']
    })

    a.process_version(v)

    assert part_year.inc.called
    assert part_month.inc.called


def test_conditional_reset_process_version_calls_reset_on_field(mocker):
    mocker.patch('punch.version_part.IntegerVersionPart.reset')
    v = ver.Version()
    part_year = vp.DateVersionPart('year', 2016, '%Y')
    part_month = vp.DateVersionPart('month', 1, '%m')
    part_build = vp.IntegerVersionPart('build')
    v.add_part(part_year)
    v.add_part(part_month)
    v.add_part(part_build)

    a = conditional_reset.ConditionalResetAction({
        'field': 'build',
        'update_fields': ['year', 'month']
    })

    a.process_version(v)

    assert part_build.reset.called


def test_conditional_reset_process_version_calls_increment_on_field(mocker):
    mocker.patch('punch.version_part.IntegerVersionPart.inc')
    strftime = mocker.patch('punch.version_part.strftime')
    strftime.return_value = '2016'
    v = ver.Version()
    part_year = vp.DateVersionPart('year', '2016', '%Y')
    part_build = vp.IntegerVersionPart('build')
    v.add_part(part_year)
    v.add_part(part_build)

    a = conditional_reset.ConditionalResetAction({
        'field': 'build',
        'update_fields': ['year']
    })

    a.process_version(v)

    assert part_build.inc.called



import subprocess
import pytest

pytestmark = pytest.mark.slow

version_file_content = """
major = '2016'
minor = '04'
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}',
}

FILES = ["README.md"]

VERSION = [
    {
        'name': 'major',
        'type': 'date',
        'fmt': '%Y'
    },
    {
        'name': 'minor',
        'type': 'date',
        'fmt': '%m'
    }
]
"""


def test_update_major(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 2016.04.")

    test_environment.ensure_file_is_present("punch_version.py",
                                            version_file_content)

    test_environment.ensure_file_is_present("punch_config.py",
                                            config_file_content)

    system_year = subprocess.check_output(['date', '+%Y'])
    system_year = system_year.decode('utf8').replace('\n', '')

    system_month = subprocess.check_output(['date', '+%m'])
    system_month = system_month.decode('utf8').replace('\n', '')

    test_environment.call(["punch", "--part", "major"])

    assert test_environment.get_file_content("README.md") == \
        "Version {}.{}.".format(system_year, system_month)


def test_update_minor(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 2016.04.")

    test_environment.ensure_file_is_present("punch_version.py",
                                            version_file_content)

    test_environment.ensure_file_is_present("punch_config.py",
                                            config_file_content)

    system_month = subprocess.check_output(['date', '+%m'])
    system_month = system_month.decode('utf8').replace('\n', '')

    test_environment.call(["punch", "--part", "minor"])

    assert test_environment.get_file_content("README.md") == \
        "Version 2016.{}.".format(system_month)



# coding: utf-8

import os
import pytest
import six
from punch import file_configuration as fc
from punch import file_updater as fu


@pytest.fixture
def temp_dir_with_version_file(temp_empty_dir):
    with open(os.path.join(temp_empty_dir, "__init__.py"), 'w') as f:
        f.write("__version__ = \"1.2.3\"")

    return temp_empty_dir


@pytest.fixture
def temp_dir_with_unicode_version_file(temp_empty_dir):
    with open(os.path.join(temp_empty_dir, "__init__.py"), 'w') as f:
        f.write("__version⚠__ = \"1.2.3\"")

    return temp_empty_dir


@pytest.fixture
def temp_dir_with_version_file_partial(temp_empty_dir):
    with open(os.path.join(temp_empty_dir, "__init__.py"), 'w') as f:
        f.write("__version__ = \"1.2\"")

    return temp_empty_dir


def test_file_updater(temp_dir_with_version_file):
    filepath = os.path.join(temp_dir_with_version_file, "__init__.py")

    current_version = {
        'major': 1,
        'minor': 2,
        'patch': 3
    }
    new_version = {
        'major': 1,
        'minor': 2,
        'patch': 4
    }

    local_variables = {
        'serializer': "__version__ = \"{{major}}.{{minor}}.{{patch}}\""
    }

    file_config = fc.FileConfiguration(filepath, local_variables)

    updater = fu.FileUpdater(file_config)
    updater.update(current_version, new_version)

    with open(filepath, 'r') as f:
        new_file_content = f.read()

    assert new_file_content == "__version__ = \"1.2.4\""


def test_file_updater_with_unicode_characters(
        temp_dir_with_unicode_version_file):
    filepath = os.path.join(temp_dir_with_unicode_version_file, "__init__.py")

    current_version = {
        'major': 1,
        'minor': 2,
        'patch': 3
    }
    new_version = {
        'major': 1,
        'minor': 2,
        'patch': 4
    }

    local_variables = {
        'serializer': "__version⚠__ = \"{{major}}.{{minor}}.{{patch}}\""
    }

    file_config = fc.FileConfiguration(filepath, local_variables)

    updater = fu.FileUpdater(file_config)
    updater.update(current_version, new_version)

    with open(filepath, 'r') as f:
        new_file_content = f.read()

    assert new_file_content == "__version⚠__ = \"1.2.4\""


def test_file_updater_with_partial_serializer(
        temp_dir_with_version_file_partial):
    filepath = os.path.join(temp_dir_with_version_file_partial, "__init__.py")

    current_version = {
        'major': 1,
        'minor': 2,
        'patch': 3
    }
    new_version = {
        'major': 1,
        'minor': 3,
        'patch': 0
    }

    local_variables = {
        'serializer': "__version__ = \"{{major}}.{{minor}}\""
    }

    file_config = fc.FileConfiguration(filepath, local_variables)

    updater = fu.FileUpdater(file_config)
    updater.update(current_version, new_version)

    with open(filepath, 'r') as f:
        new_file_content = f.read()

    assert new_file_content == "__version__ = \"1.3\""


def test_file_updater_with_nonexisting_file(temp_empty_dir):
    filepath = os.path.join(temp_empty_dir, "__init__.py")
    local_variables = {
        'serializer': "__version__ = \"{{major}}.{{minor}}\""
    }

    file_config = fc.FileConfiguration(filepath, local_variables)

    current_version = {
        'major': 1,
        'minor': 2,
        'patch': 3
    }
    new_version = {
        'major': 1,
        'minor': 3,
        'patch': 0
    }

    if six.PY2:
        expected_exception = IOError
    else:
        expected_exception = FileNotFoundError

    with pytest.raises(expected_exception) as exc:
        updater = fu.FileUpdater(file_config)
        updater.update(current_version, new_version)

    assert str(exc.value) == "The file {} does not exist".format(
        file_config.path)


def test_file_updater_preview(temp_empty_dir):
    filepath = os.path.join(temp_empty_dir, "__init__.py")
    local_variables = {
        'serializer': "__version__ = \"{{major}}.{{minor}}\""
    }

    file_config = fc.FileConfiguration(filepath, local_variables)

    current_version = {
        'major': 1,
        'minor': 2,
        'patch': 3
    }
    new_version = {
        'major': 1,
        'minor': 3,
        'patch': 0
    }

    updater = fu.FileUpdater(file_config)
    summary = updater.get_summary(current_version, new_version)

    assert summary == [("__version__ = \"1.2\"", "__version__ = \"1.3\"")]



import pytest

from punch import file_configuration as fc


@pytest.fixture
def global_variables():
    return {
        'serializer': '{{ major }}.{{ minor }}.{{ patch }}',
        'mark': 'just a mark'
    }


@pytest.fixture
def local_variables():
    return {
        'serializer': '{{ major }}.{{ minor }}'
    }


@pytest.fixture
def file_configuration_dict():
    return {
        'path': 'pkg/__init__.py',
        'serializer': '{{ major }}.{{ minor }}'
    }


def test_file_configuration_from_string_local_variables_take_precedence(
        local_variables, global_variables):
    fconf = fc.FileConfiguration(
        'pkg/__init__.py',
        local_variables,
        global_variables
    )

    assert fconf.path == 'pkg/__init__.py'
    assert fconf.config['serializer'] == '{{ major }}.{{ minor }}'
    assert fconf.config['mark'] == 'just a mark'


def test_file_configuration_from_string_can_include_global_variables(
        global_variables):
    local_variables = {
        'serializer': '__version__ = {{GLOBALS.serializer}}'
    }
    fconf = fc.FileConfiguration(
        'pkg/__init__.py',
        local_variables,
        global_variables
    )

    assert fconf.path == 'pkg/__init__.py'
    assert fconf.config['serializer'] == \
        '__version__ = {{ major }}.{{ minor }}.{{ patch }}'
    assert fconf.config['mark'] == 'just a mark'


def test_file_conf_fr_str_path_cannot_be_overridden_by_global_variables(
        local_variables, global_variables):
    global_variables['path'] = 'a/new/path'
    fconf = fc.FileConfiguration(
        'pkg/__init__.py',
        local_variables,
        global_variables
    )

    assert fconf.path == 'pkg/__init__.py'


def test_file_conf_fr_str_path_cannot_be_overridden_by_local_variables(
        local_variables, global_variables):
    local_variables['path'] = 'a/new/path'
    fconf = fc.FileConfiguration(
        'pkg/__init__.py',
        local_variables,
        global_variables
    )

    assert fconf.path == 'pkg/__init__.py'


def test_file_configuration_from_dict_local_variables_take_precedence(
        file_configuration_dict, global_variables):
    fconf = fc.FileConfiguration.from_dict(
        file_configuration_dict,
        global_variables
    )

    assert fconf.path == 'pkg/__init__.py'
    assert fconf.config['serializer'] == '{{ major }}.{{ minor }}'
    assert fconf.config['mark'] == 'just a mark'


def test_file_conf_fr_dict_path_cannot_be_overridden_by_global_variables(
        file_configuration_dict, global_variables):
    global_variables['path'] = 'a/new/path'
    fconf = fc.FileConfiguration.from_dict(
        file_configuration_dict,
        global_variables
    )

    assert fconf.path == 'pkg/__init__.py'



import mock
from punch.vcs_use_cases import release as rel


def test_pre_start_release():
    repo = mock.Mock()
    use_case = rel.VCSReleaseUseCase(repo)

    use_case.pre_start_release()

    assert repo.pre_start_release.called
    assert repo.pre_start_release.called_with()


def test_pre_start_release_can_be_called_without_release_name():
    repo = mock.Mock()
    use_case = rel.VCSReleaseUseCase(repo)

    use_case.pre_start_release()

    assert repo.pre_start_release.called
    assert repo.pre_start_release.called_with()


def test_start_release():
    repo = mock.Mock()
    use_case = rel.VCSReleaseUseCase(repo)

    use_case.start_release()

    assert repo.start_release.called
    assert repo.start_release.called_with()


def test_finish_release():
    repo = mock.Mock()
    use_case = rel.VCSReleaseUseCase(repo)

    use_case.finish_release()

    assert repo.finish_release.called
    assert repo.finish_release.called_with()


def test_post_finish_release():
    repo = mock.Mock()
    use_case = rel.VCSReleaseUseCase(repo)

    use_case.post_finish_release()

    assert repo.post_finish_release.called
    assert repo.post_finish_release.called_with()



import subprocess
import pytest

pytestmark = pytest.mark.slow

version_file_content = """
date = '20160415'
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{date}}',
}

FILES = ["README.md"]

VERSION = [
    {
        'name': 'date',
        'type': 'date',
        'fmt': '%Y%m%d'
    }
]
"""


def test_update_date(test_environment, mocker):
    test_environment.ensure_file_is_present("README.md", "Version 20160415.")

    test_environment.ensure_file_is_present("punch_version.py",
                                            version_file_content)

    test_environment.ensure_file_is_present("punch_config.py",
                                            config_file_content)

    system_date = subprocess.check_output(['date', '+%Y%m%d'])

    system_date = system_date.decode('utf8').replace('\n', '')

    test_environment.call(["punch", "--part", "date"])

    assert test_environment.get_file_content("README.md") == \
        "Version {}.".format(system_date)



# -*- coding: utf-8 -*-

from punch import action_register


def test_action_resgister():
    class ASimpleTestAction(object):
        pass

    action_register.ActionRegister.register(ASimpleTestAction, 'a_simple_test')

    assert action_register.ActionRegister.get(
        'a_simple_test') == ASimpleTestAction



from __future__ import print_function, absolute_import, division

import subprocess

import six
from punch.vcs_repositories.exceptions import RepositorySystemError


class VCSRepo(object):

    def __init__(self, working_path, config_obj):
        self.working_path = working_path
        self.config_obj = config_obj

        self._check_config()
        self._set_command()
        self._check_system()

    def _check_config(self):
        pass

    def _set_command(self):
        self.commands = [None]
        self.command = None

    def _check_system(self):
        null_commands = self.commands + ["--help"]

        if six.PY2:
            not_found_exception = IOError
        else:
            not_found_exception = FileNotFoundError

        try:
            if six.PY2:
                subprocess.check_output(null_commands)
            else:
                subprocess.check_call(null_commands, stdout=subprocess.DEVNULL)

        except not_found_exception:
            raise RepositorySystemError("Cannot run {}".format(self.command))
        except subprocess.CalledProcessError:
            raise RepositorySystemError(
                "Error running {}".format(self.command))

    def _run(self, command_line, error_message=None):
        p = subprocess.Popen(
            command_line,
            cwd=self.working_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        stdout, stderr = p.communicate()

        if p.returncode != 0:
            if error_message is not None:
                raise RepositorySystemError(error_message.format(stderr))
            else:
                error_text = "An error occurred executing" + \
                    " '{}': {}\nProcess output was: {}"
                error_message = error_text.format(
                    " ".join(command_line),
                    stderr.decode('utf8'), stdout.decode('utf8')
                )
                raise RepositorySystemError(error_message)

        return stdout.decode('utf8')

    def pre_start_release(self):
        pass

    def start_release(self):
        pass

    def finish_release(self):
        pass

    def post_finish_release(self):
        pass



from __future__ import print_function, absolute_import, division


class RepositoryError(Exception):
    pass


class RepositoryWorkflowError(RepositoryError):
    pass


class RepositorySystemError(RepositoryError):
    pass


class RepositoryConfigurationError(RepositoryError):
    pass


class RepositoryStatusError(RepositoryError):
    pass



import pytest

pytestmark = pytest.mark.slow


def test_punch_version_flag(test_environment):
    # output = test_environment_fixture.output(["punch", "--version"])

    # Punch version 1.0.1
    # Copyright (C) 2016 Leonardo Giordani

    expected_output = """
    Copyright \(C\) \d{4} Leonardo Giordani
    This is free software, see the LICENSE file.
    Source: https://github.com/lgiordani/punch
    Documentation: http://punch.readthedocs.io/en/latest/
    """

    test_environment.compare_output(expected_output, ["punch", "--version"])



from __future__ import print_function, absolute_import, division


class ConditionalResetAction(object):

    def __init__(self, adict):
        self.field = adict['field']
        self.update_fields = adict.get('update_fields', None)

    def process_version(self, version):
        new_version = version.copy()

        reset_part = new_version.get_part(self.field)

        for f in self.update_fields:
            update_part = new_version.get_part(f)
            update_part.inc()

        if new_version == version:
            reset_part.inc()
        else:
            reset_part.reset()

        return new_version



from __future__ import print_function, absolute_import, division


class SetPartAction(object):

    def __init__(self, adict):
        self.parts = adict

    def process_version(self, version):
        new_version = version.copy()
        new_version.set(self.parts)

        return new_version



import pytest

from punch import vcs_configuration as vc


@pytest.fixture
def global_variables():
    return {
        'serializer': '{{ major }}.{{ minor }}.{{ patch }}',
        'mark': 'just a mark'
    }


@pytest.fixture
def vcs_configuration_dict():
    return {
        'name': 'git',
        'commit_message': "Version updated to {{ new_version }}",
        'finish_release': True,
        'options': {
            'make_release_branch': False,
            'annotate_tags': False,
            'annotation_message': '',
        }
    }


@pytest.fixture
def special_variables():
    return {
        'current_version': '1.2.3',
        'new_version': '1.3.0'
    }


def test_vcs_configuration_from_string(
        vcs_configuration_dict, global_variables, special_variables):
    vcsconf = vc.VCSConfiguration(vcs_configuration_dict['name'],
                                  vcs_configuration_dict['options'],
                                  global_variables,
                                  special_variables,
                                  vcs_configuration_dict['commit_message']
                                  )

    expected_options = {
        'make_release_branch': False,
        'annotate_tags': False,
        'annotation_message': '',
        'current_version': '1.2.3',
        'new_version': '1.3.0'
    }

    assert vcsconf.name == 'git'
    assert vcsconf.commit_message == "Version updated to 1.3.0"
    assert vcsconf.finish_release is True
    assert vcsconf.options == expected_options


def test_vcs_configuration_from_dict(
        vcs_configuration_dict, global_variables, special_variables):
    vcsconf = vc.VCSConfiguration.from_dict(
        vcs_configuration_dict,
        global_variables,
        special_variables
    )

    expected_options = {
        'make_release_branch': False,
        'annotate_tags': False,
        'annotation_message': '',
        'current_version': '1.2.3',
        'new_version': '1.3.0'
    }

    assert vcsconf.name == 'git'
    assert vcsconf.commit_message == "Version updated to 1.3.0"
    assert vcsconf.finish_release is True
    assert vcsconf.options == expected_options


def test_vcs_configuration_from_dict_without_commit_message(
        vcs_configuration_dict, global_variables, special_variables):
    vcs_configuration_dict.pop('commit_message')
    vcsconf = vc.VCSConfiguration.from_dict(
        vcs_configuration_dict,
        global_variables,
        special_variables
    )

    expected_options = {
        'make_release_branch': False,
        'annotate_tags': False,
        'annotation_message': '',
        'current_version': '1.2.3',
        'new_version': '1.3.0'
    }

    assert vcsconf.name == 'git'
    assert vcsconf.commit_message == "Version updated 1.2.3 -> 1.3.0"
    assert vcsconf.finish_release is True
    assert vcsconf.options == expected_options


def test_vcs_configuration_from_dict_without_finish_release(
        vcs_configuration_dict, global_variables, special_variables):
    vcs_configuration_dict.pop('finish_release')
    vcsconf = vc.VCSConfiguration.from_dict(
        vcs_configuration_dict,
        global_variables,
        special_variables
    )

    expected_options = {
        'make_release_branch': False,
        'annotate_tags': False,
        'annotation_message': '',
        'current_version': '1.2.3',
        'new_version': '1.3.0'
    }

    assert vcsconf.name == 'git'
    assert vcsconf.commit_message == "Version updated to 1.3.0"
    assert vcsconf.finish_release is True
    assert vcsconf.options == expected_options


def test_vcs_configuration_from_dict_without_options(
        vcs_configuration_dict, global_variables, special_variables):
    vcs_configuration_dict.pop('options')
    vcsconf = vc.VCSConfiguration.from_dict(
        vcs_configuration_dict,
        global_variables,
        special_variables
    )

    assert vcsconf.name == 'git'
    assert vcsconf.commit_message == "Version updated to 1.3.0"
    assert vcsconf.finish_release is True


def test_vcs_configuration_from_dict_can_use_global_variables(
        vcs_configuration_dict, global_variables, special_variables):
    vcs_configuration_dict['commit_message'] = "Mark: {{ mark }}"

    vcsconf = vc.VCSConfiguration.from_dict(
        vcs_configuration_dict,
        global_variables,
        special_variables
    )

    assert vcsconf.commit_message == "Mark: just a mark"


def test_vcs_configuration_from_dict_special_variables_take_precedence(
        vcs_configuration_dict, global_variables, special_variables):
    vcs_configuration_dict['commit_message'] = "{{ current_version }}"
    global_variables['current_version'] = "5.0.0"

    vcsconf = vc.VCSConfiguration.from_dict(
        vcs_configuration_dict,
        global_variables,
        special_variables
    )

    assert vcsconf.commit_message == "1.2.3"


def test_vcs_configuration_from_dict_options_templates_are_processed(
        vcs_configuration_dict, global_variables, special_variables):
    vcs_configuration_dict['options']['annotation_message'] = \
        "Updated {{ current_version}} -> {{ new_version }}"
    vcsconf = vc.VCSConfiguration.from_dict(
        vcs_configuration_dict,
        global_variables,
        special_variables
    )

    expected_options = {
        'make_release_branch': False,
        'annotate_tags': False,
        'annotation_message': 'Updated 1.2.3 -> 1.3.0',
        'current_version': '1.2.3',
        'new_version': '1.3.0'
    }

    assert vcsconf.options == expected_options



from __future__ import print_function, absolute_import, division

import collections
from punch import file_configuration as fc
from punch.helpers import import_file


class ConfigurationVersionError(Exception):
    "An exception used to signal that the configuration file version is wrong"


class PunchConfig(object):

    def __init__(self, config_filepath):

        configuration_module = import_file(config_filepath)

        try:
            self.__config_version__ = configuration_module.__config_version__
        except AttributeError:
            raise ValueError(
                "Given config file is invalid:" +
                " missing '__config_version__' variable"
            )

        if configuration_module.__config_version__ > 1:
            raise ConfigurationVersionError(
                "Unsupported configuration file version" +
                " {}".format(configuration_module.__config_version__)
            )

        try:
            self.globals = configuration_module.GLOBALS
        except AttributeError:
            self.globals = {}

        try:
            files = configuration_module.FILES
        except AttributeError:
            raise ValueError(
                "Given config file is invalid: missing 'FILES' attribute")

        self.files = []
        for file_configuration in files:
            if isinstance(file_configuration, collections.Mapping):
                self.files.append(fc.FileConfiguration.from_dict(
                    file_configuration, self.globals))
            else:
                self.files.append(fc.FileConfiguration(
                    file_configuration, {}, self.globals))

        try:
            self.version = configuration_module.VERSION
        except AttributeError:
            raise ValueError(
                "Given config file is invalid: missing 'VERSION' attribute")

        try:
            self.vcs = configuration_module.VCS
            if 'name' not in self.vcs.keys():
                raise ValueError("Missing key 'name' in VCS configuration")
        except AttributeError:
            self.vcs = None

        try:
            self.actions = configuration_module.ACTIONS
        except AttributeError:
            self.actions = {}

        self.actions.update(
            {
                'punch:increase': {
                    'type': 'increase_part'
                },
                'punch:set': {
                    'type': 'set_part'
                }
            }
        )



import subprocess
import pytest

pytestmark = pytest.mark.slow

system_year = subprocess.check_output(['date', '+%Y'])
system_year = system_year.decode('utf8').replace('\n', '')

system_month = subprocess.check_output(['date', '+%m'])
system_month = system_month.decode('utf8').replace('\n', '').lstrip('0')

version_file_content = """
year = {}
month = {}
build = 8
""".format(system_year, system_month)

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{year}}.{{month}}.{{build}}',
}

FILES = ['README.md']

ACTIONS = {
    'mbuild': {
        'type': 'conditional_reset',
        'field': 'build',
        'update_fields': ['year', 'month']
    }
}

VERSION = [
    {
        'name': 'year',
        'type': 'date',
        'fmt': 'YYYY'
    },
    {
        'name': 'month',
        'type': 'date',
        'fmt': 'MM'
    },
    {
        'name': 'build',
        'type': 'integer',
        'start_value': 0
    }
]

ACTIONS = {
    'build': {
        'type': 'conditional_reset',
        'field': 'build',
        'update_fields': ['year', 'month']
    }
}
"""


def test_action_refresh(script_runner, test_environment):
    test_environment.ensure_file_is_present(
        "README.md",
        "Version {}.{}.8.".format(system_year, system_month)
    )

    test_environment.ensure_file_is_present("punch_version.py",
                                            version_file_content)

    test_environment.ensure_file_is_present("punch_config.py",
                                            config_file_content)

    ret = test_environment.call(['punch', '--action', 'build'])

    assert not ret.stderr
    assert test_environment.get_file_content("README.md") == \
        "Version {}.{}.9.".format(system_year, system_month)



import pytest

pytestmark = pytest.mark.slow

version_file_content = """
major = 1
minor = 0
patch = 0
"""

config_file_content = """
__config_version__ = 1

GLOBALS = {
    'serializer': '{{major}}.{{minor}}.{{patch}}',
}

FILES = ["README.md"]

VERSION = ['major', 'minor', 'patch']
"""


def test_use_undefined_action(test_environment):
    test_environment.ensure_file_is_present("README.md", "Version 1.0.0")

    test_environment.ensure_file_is_present(
        "punch_version.py",
        version_file_content
    )

    test_environment.ensure_file_is_present(
        "punch_config.py",
        config_file_content
    )

    out = test_environment.output([
        "punch",
        "--action", "punch:foobar",
        "--action-options", "part=major"
    ])

    assert 'KeyError: punch:foobar' not in out
    assert 'punch:foobar' in out



# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division

import sys
from datetime import datetime


def _strftime(fmt):
    return datetime.now().strftime(fmt)


def strftime(fmt):
    calver_syntax = {
        'YYYY': {
            'fmt': '%Y',
            'strip': False
        },
        'YY': {
            'fmt': '%y',
            'strip': True
        },
        '0M': {
            'fmt': '%m',
            'strip': False
        },
        '0D': {
            'fmt': '%d',
            'strip': False
        },
        'MM': {
            'fmt': '%m',
            'strip': True
        },
        'DD': {
            'fmt': '%d',
            'strip': True
        }
    }

    newfmt = calver_syntax.get(fmt, {'fmt': fmt, 'strip': False})
    value = _strftime(newfmt['fmt'])

    if newfmt['strip']:
        return value.lstrip('0')

    return value


class VersionPart(object):

    @classmethod
    def from_dict(cls, dic):
        try:
            part_type = dic.pop('type')
        except KeyError:
            part_type = 'integer'

        class_name = part_type.title().replace("_", "") + 'VersionPart'
        part_class = getattr(sys.modules[__name__], class_name)

        return part_class(**dic)


class IntegerVersionPart(VersionPart):

    def __init__(self, name, value=None, start_value=None):
        self.name = name

        if start_value is None:
            self.start_value = 0
        else:
            self.start_value = start_value

        if value is None:
            self.value = self.start_value
        else:
            self.set(value)

    def inc(self):
        self.value = self.value + 1

    def set(self, value):
        self.value = int(value)

    def reset(self):
        self.value = self.start_value

    def copy(self):
        return IntegerVersionPart(self.name, self.value, self.start_value)


class ValueListVersionPart(VersionPart):

    def __init__(self, name, value, allowed_values):
        self.name = name
        self.allowed_values = allowed_values

        if value is None:
            self.value = allowed_values[0]
        else:
            self.set(value)

        # When copying this does not take the object itself
        self.values = [v for v in allowed_values]

    def set(self, value):
        if value not in self.allowed_values:
            raise ValueError(
                "The given value {} is not allowed," +
                " the list of possible values is {}",
                value,
                self.allowed_values)
        self.value = value

    def inc(self):
        idx = self.values.index(self.value)
        self.value = self.values[(idx + 1) % len(self.values)]

    def reset(self):
        self.value = self.values[0]

    def copy(self):
        return ValueListVersionPart(self.name, self.value, self.values)


class DateVersionPart(VersionPart):

    def __init__(self, name, value, fmt):
        self.name = name
        self.fmt = fmt

        if value is None:
            self.value = strftime(fmt)
        else:
            self.value = str(value)

    def reset(self):
        self.value = strftime(self.fmt)

    def inc(self):
        self.reset()

    def copy(self):
        return DateVersionPart(self.name, self.value, self.fmt)



import subprocess

import os
import pytest
from punch import vcs_configuration as vc
from punch.vcs_repositories import git_repo as gr, exceptions as re

pytestmark = pytest.mark.slow


@pytest.fixture
def temp_empty_git_dir(temp_empty_dir):
    subprocess.check_call(["git", "init", "-q", temp_empty_dir])
    subprocess.check_call(["git", "config", "user.email",
                           "py.test@email.com"], cwd=temp_empty_dir)
    subprocess.check_call(["git", "config", "user.name",
                           "PyTest"], cwd=temp_empty_dir)

    return temp_empty_dir


@pytest.fixture
def temp_git_dir(temp_empty_git_dir, safe_devnull):
    with open(os.path.join(temp_empty_git_dir, "README.md"), "w") as f:
        f.writelines(["# Test file", "This is just a test file for punch"])

    subprocess.check_call(["git", "add", "README.md"],
                          cwd=temp_empty_git_dir, stdout=safe_devnull)
    subprocess.check_call(["git", "commit", "-m", "Initial addition"],
                          cwd=temp_empty_git_dir,
                          stdout=safe_devnull)

    return temp_empty_git_dir


@pytest.fixture
def empty_vcs_configuration():
    return vc.VCSConfiguration('git', {}, {},
                               {'current_version': 'a', 'new_version': 'b'})


def test_get_current_branch(temp_git_dir, empty_vcs_configuration):
    repo = gr.GitRepo(temp_git_dir, empty_vcs_configuration)
    assert repo.get_current_branch() == 'master'


def test_get_tags(temp_git_dir, empty_vcs_configuration):
    repo = gr.GitRepo(temp_git_dir, empty_vcs_configuration)
    assert repo.get_tags() == ''


def test_init(temp_empty_git_dir, empty_vcs_configuration):
    repo = gr.GitRepo(temp_empty_git_dir, empty_vcs_configuration)

    assert repo.working_path == temp_empty_git_dir


def test_init_with_uninitialized_dir(temp_empty_dir, empty_vcs_configuration):
    with pytest.raises(re.RepositorySystemError) as exc:
        gr.GitRepo(temp_empty_dir, empty_vcs_configuration)

    assert str(exc.value) == \
        "The current directory {} is not a Git repository".format(
            temp_empty_dir)


def test_pre_start_release(temp_git_dir, empty_vcs_configuration):
    repo = gr.GitRepo(temp_git_dir, empty_vcs_configuration)
    repo.pre_start_release()

    assert repo.get_current_branch() == 'master'


def test_pre_start_release_starting_from_different_branch(
        temp_git_dir, safe_devnull, empty_vcs_configuration):
    subprocess.check_call(
        ["git", "checkout", "-b", "new_branch"],
        cwd=temp_git_dir,
        stdout=safe_devnull,
        stderr=safe_devnull
    )

    repo = gr.GitRepo(temp_git_dir, empty_vcs_configuration)
    repo.pre_start_release()

    assert repo.get_current_branch() == 'master'


def test_pre_start_release_starting_from_different_branch_w_unstaged_changes(
        temp_git_dir, safe_devnull, empty_vcs_configuration):
    subprocess.check_call(
        ["git", "checkout", "-b", "new_branch"],
        cwd=temp_git_dir,
        stdout=safe_devnull,
        stderr=safe_devnull
    )
    with open(os.path.join(temp_git_dir, "README.md"), "w") as f:
        f.writelines(["Unstaged lines"])

    repo = gr.GitRepo(temp_git_dir, empty_vcs_configuration)
    repo.pre_start_release()

    assert repo.get_current_branch() == 'master'


def test_pre_start_release_starting_from_different_branch_w_uncom_changes(
        temp_git_dir, safe_devnull, empty_vcs_configuration):
    subprocess.check_call(
        ["git", "checkout", "-b", "new_branch"],
        cwd=temp_git_dir,
        stdout=safe_devnull,
        stderr=safe_devnull
    )
    with open(os.path.join(temp_git_dir, "README.md"), "w") as f:
        f.writelines(["Unstaged lines"])
    subprocess.check_call(
        ["git", "add", "README.md"],
        cwd=temp_git_dir,
        stdout=safe_devnull,
        stderr=safe_devnull
    )

    repo = gr.GitRepo(temp_git_dir, empty_vcs_configuration)
    with pytest.raises(re.RepositoryStatusError):
        repo.pre_start_release()


def test_start_release(temp_git_dir, empty_vcs_configuration):
    repo = gr.GitRepo(temp_git_dir, empty_vcs_configuration)
    repo.pre_start_release()
    repo.start_release()
    assert repo.get_current_branch() == "b"


def test_finish_release_without_changes(temp_git_dir, empty_vcs_configuration):
    repo = gr.GitRepo(temp_git_dir, empty_vcs_configuration)
    repo.pre_start_release()
    repo.start_release()
    assert 'b' in repo.get_branches()

    repo.finish_release()
    assert repo.get_current_branch() == "master"
    assert 'b' not in repo.get_branches()
    assert 'b' not in repo.get_tags()


def test_finish_release_with_message(temp_git_dir, safe_devnull):
    release_name = "1.0"
    commit_message = "A commit message"
    config = vc.VCSConfiguration(
        'git', {}, global_variables={},
        special_variables={'new_version': release_name},
        commit_message=commit_message
    )

    repo = gr.GitRepo(temp_git_dir, config)
    repo.pre_start_release()
    repo.start_release()
    assert release_name in repo.get_branches()

    with open(os.path.join(temp_git_dir, "version.txt"), "w") as f:
        f.writelines([release_name])

    subprocess.check_call(
        ["git", "add", "version.txt"],
        cwd=temp_git_dir,
        stdout=safe_devnull,
        stderr=safe_devnull
    )

    repo.finish_release()

    p = subprocess.Popen(["git", "log"], cwd=temp_git_dir,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    assert commit_message in stdout.decode('utf8')

    assert release_name in repo.get_tags()
    assert release_name not in repo.get_branches()


def test_release_without_release_branch(temp_git_dir, safe_devnull):
    release_name = "1.0"
    commit_message = "A commit message"
    config = vc.VCSConfiguration(
        'git', {'make_release_branch': False}, global_variables={},
        special_variables={'new_version': release_name},
        commit_message=commit_message
    )

    repo = gr.GitRepo(temp_git_dir, config)
    repo.pre_start_release()
    repo.start_release()
    assert release_name not in repo.get_branches()

    with open(os.path.join(temp_git_dir, "version.txt"), "w") as f:
        f.writelines([release_name])

    subprocess.check_call(
        ["git", "add", "version.txt"],
        cwd=temp_git_dir,
        stdout=safe_devnull,
        stderr=safe_devnull
    )

    repo.finish_release()

    p = subprocess.Popen(["git", "log"], cwd=temp_git_dir,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    assert commit_message in stdout.decode('utf8')

    assert release_name in repo.get_tags()
    assert release_name not in repo.get_branches()


def test_release_with_explicit_release_branch(temp_git_dir, safe_devnull):
    release_name = "1.0"
    commit_message = "A commit message"
    config = vc.VCSConfiguration(
        'git', {'make_release_branch': True}, global_variables={},
        special_variables={'new_version': release_name},
        commit_message=commit_message
    )

    repo = gr.GitRepo(temp_git_dir, config)
    repo.pre_start_release()
    repo.start_release()
    assert release_name in repo.get_branches()

    with open(os.path.join(temp_git_dir, "version.txt"), "w") as f:
        f.writelines([release_name])

    subprocess.check_call(
        ["git", "add", "version.txt"],
        cwd=temp_git_dir,
        stdout=safe_devnull,
        stderr=safe_devnull
    )

    repo.finish_release()

    p = subprocess.Popen(["git", "log"], cwd=temp_git_dir,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    assert commit_message in stdout.decode('utf8')

    assert release_name in repo.get_tags()
    assert release_name not in repo.get_branches()


def test_finish_release_with_custom_tag(temp_git_dir, safe_devnull):
    release_name = "1.0"
    commit_message = "A commit message"
    tag = "Version_{}".format(release_name)

    config = vc.VCSConfiguration(
        'git', {'tag': tag}, global_variables={},
        special_variables={'new_version': release_name},
        commit_message=commit_message
    )

    repo = gr.GitRepo(temp_git_dir, config)
    repo.pre_start_release()
    repo.start_release()

    with open(os.path.join(temp_git_dir, "version.txt"), "w") as f:
        f.writelines([release_name])

    subprocess.check_call(
        ["git", "add", "version.txt"],
        cwd=temp_git_dir,
        stdout=safe_devnull,
        stderr=safe_devnull
    )

    repo.finish_release()

    assert tag in repo.get_tags()


def test_finish_release_custom_tag_cannot_contain_spaces(temp_git_dir):
    release_name = "1.0"
    commit_message = "A commit message"
    tag = "Version {}".format(release_name)

    config = vc.VCSConfiguration(
        'git', {'tag': tag}, global_variables={},
        special_variables={'new_version': release_name},
        commit_message=commit_message
    )

    with pytest.raises(re.RepositoryConfigurationError):
        gr.GitRepo(temp_git_dir, config)


def test_finish_release_with_annotated_tag(temp_git_dir, safe_devnull):
    release_name = "1.0"
    commit_message = "A commit message"
    annotation_message = "An annotation message"

    config = vc.VCSConfiguration(
        'git',
        {'annotate_tags': True,
         'annotation_message': annotation_message},
        global_variables={},
        special_variables={'new_version': release_name},
        commit_message=commit_message
    )

    repo = gr.GitRepo(temp_git_dir, config)
    repo.pre_start_release()
    repo.start_release()

    with open(os.path.join(temp_git_dir, "version.txt"), "w") as f:
        f.writelines([release_name])

    subprocess.check_call(
        ["git", "add", "version.txt"],
        cwd=temp_git_dir,
        stdout=safe_devnull,
        stderr=safe_devnull
    )

    repo.finish_release()

    p = subprocess.Popen(["git", "log"], cwd=temp_git_dir,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    assert commit_message in stdout.decode('utf8')

    p = subprocess.Popen(
        ["git", "show", release_name],
        cwd=temp_git_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    stdout, stderr = p.communicate()
    assert annotation_message in stdout.decode('utf8')

    assert release_name in repo.get_tags()
    assert release_name not in repo.get_branches()


def test_tag(temp_git_dir, empty_vcs_configuration):
    repo = gr.GitRepo(temp_git_dir, empty_vcs_configuration)

    repo.tag("just_a_tag")

    assert "just_a_tag" in repo.get_tags()



from __future__ import print_function, absolute_import, division

import os

import re
import six

from punch.vcs_repositories import vcs_repo as vr
from punch.vcs_repositories import exceptions as e


class HgRepo(vr.VCSRepo):
    DEFAULT_BRANCH = 'default'

    def __init__(self, working_path, config_obj):
        if six.PY2:
            super(HgRepo, self).__init__(working_path, config_obj)
        else:
            super().__init__(working_path, config_obj)

        self.branch = self.config_obj.options.get('branch', 'default')
        self._recorded_branch = None

    def get_current_branch(self):
        stdout = self._run([self.command, "branch"])

        branch = stdout.replace("\n", "")

        return branch

    def get_branches(self):
        stdout = self._run([self.command, "branches"])
        return {self._parse_branch_line(l) for l in stdout.splitlines()}

    def get_tags(self):
        tags_str = self._run([self.command, "tags"])
        tags = map(
            lambda l: l.rsplit(" ", 1)[0].strip(),
            tags_str.splitlines()
        )
        return "\n".join(tags)

    def get_summary(self):
        output = self._run([self.command, "summary"])
        keys = {"branch", "commit", "update"}
        summary = {}
        for l in output.splitlines():
            try:
                k, body = l.split(": ", 1)
                if k in keys:
                    summary[k] = body
            except ValueError:
                pass

        return summary

    def pre_start_release(self):
        if not self._is_clean():
            raise e.RepositoryStatusError(
                "Cannot update default while repository" +
                " contains uncommitted changes")
        self._recorded_branch = self.get_current_branch()

        self._change_branch(self.branch)

        branch = self.get_current_branch()

        if branch != self.branch:
            raise e.RepositoryStatusError(
                "Current branch shall be {} but is {}".format(
                    self.branch, branch))

    def start_release(self):
        pass

    def finish_release(self):
        self.get_current_branch()
        try:
            if self._is_clean():
                return
            command_line = [self.command, "commit"]
            command_line.extend(["-m", self.config_obj.commit_message])
            self._run(command_line)
            tag = self._configured_tag()
            self.tag(tag)
        finally:
            self._recover_branch()

    def tag(self, tag):
        self._run([self.command, "tag", tag])

    def _recover_branch(self):
        if self._recorded_branch is not None:
            self._change_branch(self._recorded_branch)
            self._recorded_branch = None

    def _change_branch(self, branch):
        self._run([self.command, "update", branch])

    def _check_config(self):
        # Tag names cannot contain spaces
        tag = self.config_obj.options.get('tag', '')
        if ' ' in tag:
            raise e.RepositoryConfigurationError(
                """You specified "'tag': {}".""".format(tag) +
                " Tag names cannot contain spaces")
        if re.match("^\d+$", tag):
            raise e.RepositoryConfigurationError(
                """You specified "'tag': {}".""".format(tag) +
                " Tag names cannot be just digits")

    def _check_system(self):
        if six.PY2:
            super(HgRepo, self)._check_system()
        else:
            super()._check_system()

        if not os.path.exists(os.path.join(self.working_path, '.hg')):
            raise e.RepositorySystemError(
                "The current directory {} is not a Hg repository".format(
                    self.working_path))

    def _set_command(self):
        self.commands = ['hg']
        self.command = 'hg'

    def _is_clean(self):
        return self.get_summary()["commit"].endswith("(clean)")

    @classmethod
    def _parse_branch_line(cls, line):
        return re.match("(?P<tag>.+)\s+\d+:.+$", line).group("tag").strip()

    def _configured_tag(self):
        try:
            return self.config_obj.options['tag']
        except KeyError:
            return self.config_obj.options['new_version']
